from setuptools import setup
from pathlib import Path
from Cython.Build import cythonize

setup(
    long_description=Path('README.md').read_text(encoding='utf8'),
    long_description_content_type='text/markdown',
    extras_require={
        'gui': [
            "pygame>=2.0.1",
        ]
    },
    py_modules=['pyfxr', 'pyfxr_gui'],
    entry_points={
        'console_scripts': [
            'pyfxr = pyfxr_gui:main [gui]',
        ]
    },
    ext_modules=cythonize(
        "_pyfxr.pyx",
        compiler_directives={'embedsignature': True}
    ),
    zip_safe=False,
)



from math import sin, pi, copysign, cos
import random
import abc
import sys
from functools import lru_cache
import inspect

try:
    import pygame
except ModuleNotFoundError:
    sys.exit(
        "Pygame was not found. To use the GUI, make sure you install "
        "pyfxr with\n\npip install pyfxr[gui]"
    )
import pygame.draw
import pygame.mixer
from pygame import Rect
import time

import pyfxr


BACKGROUND = pygame.Color((200, 128, 50))
WHITE_KEY = pygame.Color((200, 200, 180))
BLACK_KEY = pygame.Color((50, 50, 50))
screen: pygame.Surface = None


def lighter(color):
    return pygame.Color(
        128 + color.r >> 1,
        128 + color.g >> 1,
        128 + color.b >> 1,
    )

def darker(color):
    return pygame.Color(
        color.r >> 1,
        color.g >> 1,
        color.b >> 1,
    )


def rounded_rect(color, rect, radius=5):
    highlight = darker(color)
    pygame.draw.rect(screen, color, rect, width=0, border_radius=radius)
    pygame.draw.rect(screen, highlight, rect, width=1, border_radius=radius)


def text(text, pos, color='black', align='left'):
    surf = text_surf(text, color)
    x, y = pos
    w, h = surf.get_size()
    if align == 'left':
        pass
    elif align == 'center':
        x -= w // 2
    elif align == 'right':
        x -= w
    else:
        raise ValueError(
            "align must be left, right, or center"
        )
    screen.blit(surf, (x, y))
    return pygame.Rect(x, y, *surf.get_size())


@lru_cache()
def text_surf(text, color='black'):
    """Get a rendered text surface, with caching."""
    if isinstance(color, str):
        color = pygame.Color(color)
    return font.render(text, True, color)


def draw():
    screen.fill(BACKGROUND)

    for w in WIDGETS:
        w.draw()


class Button(abc.ABC):
    def __init__(self, rect, color, radius=5):
        self.rect = rect
        self.color = color
        self.radius = radius
        self.selected = False

    def draw(self):
        color = lighter(self.color) if self.selected else self.color
        rounded_rect(color, self.rect)

    def click(self):
        self.selected = True
        self.on_click()

    def release(self):
        self.selected = False

    @abc.abstractmethod
    def on_click(self):
        """Implement this to define click interactions."""


class Key(Button):
    def __init__(self, note, *args, **kwargs):
        self.note = note
        super().__init__(*args, **kwargs)

    def on_click(self):
        print(
            "tone = pyfxr.tone(",
            f"    {self.note!r}",
            "    attack=0.05,",
            "    sustain=0.0,",
            "    release=0.1,",
            "    wavetable=wavetable,",
            ")",
            sep="\n"
        )
        s = pygame.mixer.Sound(
            buffer=pyfxr.tone(
                self.note,
                attack=0.05,
                sustain=0.0,
                release=0.1,
                wavetable=Waveform.current
            )
        )
        s.set_volume(0.5)
        s.play()


class Label(Button):
    def __init__(self, text, pos, color='black', align='left'):
        self.color = color
        self.align = align
        self.surf = None
        self._text = text
        self.rect = Rect(-1, -1, 0, 0)
        self.pos = pos

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        self._text = value
        self.surf = None
        self.rect = Rect(-1, -1, 0, 0)

    def draw(self):
        self.rect = text(self._text, self.pos, self.color, self.align)

    def on_click(self):
        """Labels are not click-sensitive so this does nothing."""


class Waveform:
    current = None

    def __init__(self, waveform, rect):
        self.waveform_str = inspect.cleandoc(waveform)
        self.waveform = eval(waveform)
        self.rect = rect
        wf = memoryview(self.waveform)
        points = []
        for i in range(0, 1024, 8):
            x = self.rect.left + i / 1024 * self.rect.width
            y = self.rect.centery + wf[i] / (1 << 15) * (self.rect.height / 2.6)
            points.append((x, y))
        self.points = points

        if Waveform.current is None:
            Waveform.current = self.waveform

    def draw(self):
        if self.current is self.waveform:
            rounded_rect(pygame.Color('white'), self.rect)
        else:
            rounded_rect(WHITE_KEY, self.rect)
        pygame.draw.aalines(
            screen,
            'red',
            False,
            self.points
        )

    def click(self):
        print(f"wavetable = {self.waveform_str}")
        Waveform.current = self.waveform

    def release(self):
        pass


WIDGETS = []


def make_keyboard():
    notes = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
    notes = [f'{note}{octave}' for octave in (3, 4, 5) for note in notes][:25]

    keyboard = Rect(0, 400, 800, 200)
    x = keyboard.left
    white_notes = []
    black_notes = []

    n_white_keys = sum('#' not in note for note in notes)
    key_width = keyboard.width / n_white_keys
    black_width = 2 * (key_width * 2 // 6)

    for note in notes:
        if '#' in note:
            black_notes.append(Key(
                note,
                Rect(x - black_width // 2, keyboard.top - 3, black_width, keyboard.height * 2 // 3),
                BLACK_KEY,
                radius=3
            ))
        else:
            white_notes.append(Key(
                note,
                Rect(int(x), keyboard.top, int(x + key_width) - int(x) + 1, keyboard.height),
                WHITE_KEY,
            ))
            x += key_width
    WIDGETS.extend(white_notes)
    WIDGETS.extend(black_notes)

    waves = [
        "pyfxr.Wavetable.sine()",
        "pyfxr.Wavetable.square()",
        "pyfxr.Wavetable.saw()",
        "pyfxr.Wavetable.triangle()",
        """pyfxr.Wavetable.from_function(
            lambda t: sin(t) * 0.7 + 0.3 * sin(6 * t + 0.5)
        )""",
        """pyfxr.Wavetable.from_function(
            lambda t: sin(t) * 0.5 + 0.5 * sin(2 * t)
        )""",
        """pyfxr.Wavetable.from_function(
            lambda t: sin(t) * 0.4 + 0.3 * sin(2 * t) + 0.3 * sin(10 * t - 1) * sin(t / 2)
        )""",
        """pyfxr.Wavetable.from_function(
            lambda t: 1 - 2 * abs(sin(t / 2 + pi / 2))
        )""",
        """pyfxr.Wavetable.from_function(
            lambda t: sin(10 * t) * sin(t / 2)
        )""",
        "pyfxr.Wavetable.square(0.8)",
        """pyfxr.Wavetable.from_function(
            lambda t: (sin(t) + sin(3 * t) + sin(5 * t)) / 3
        )""",
        """pyfxr.Wavetable.from_function(
            lambda t: sin(t) * 0.9 + 0.1 * copysign(sin(t), sin(3 * t))
        )""",
    ]

    padding = 10
    grid = Rect(30, 60, 800 - 60, 300)
    for i, w in enumerate(waves):
        ycell, xcell = divmod(i, 4)
        width = (grid.width - 3 * padding) / 4
        height = (grid.height - 3 * padding) / 3
        x = grid.left + xcell * (width + padding)
        y = grid.top + ycell * (height + padding)
        WIDGETS.append(Waveform(w, Rect(x, y, width, height)))

    WIDGETS.append(Label("Example Waveforms", (30, 40)))


make_keyboard()


def widget_at(pos):
    for b in reversed(WIDGETS):
        if b.rect.collidepoint(pos):
            return b
    return None


def main():
    global screen, font
    pygame.mixer.pre_init(
        44100,
        channels=1
    )
    pygame.init()
    pygame.display.set_caption("pyfxr")
    screen = pygame.display.set_mode((800, 600))
    font = pygame.font.SysFont('sans-serif', 24, bold=False)

    while True:
        draw()
        pygame.display.flip()

        ev = pygame.event.wait()
        if ev.type == pygame.QUIT:
            return
        elif ev.type == pygame.MOUSEBUTTONDOWN:
            if ev.button == pygame.BUTTON_LEFT:
                clicked = widget_at(ev.pos)
                if clicked:
                    clicked.click()
        elif ev.type == pygame.MOUSEBUTTONUP:
            if ev.button == pygame.BUTTON_LEFT:
                if clicked:
                    clicked.release()
                    clicked = None


def play():
    sine = pyfxr.Waveform.square()

    start = time.perf_counter_ns()
    tone = pyfxr.pluck(duration=1.0, pitch='A4')
    end = time.perf_counter_ns()

    print("Generated tone in", (end - start) // 1000, "us")

    pygame.mixer.Sound(buffer=tone).play()

    time.sleep(tone.duration)

    tone.save('output.wav')


if __name__ == '__main__':
    main()



#cython: language_level=3

from libc.stdint cimport int16_t, uint32_t, uint64_t
from libc.math cimport sin, pi, floor
from libc.stdlib cimport rand, abs

cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cpython.buffer cimport PyObject_GetBuffer


cdef float AMPLITUDE = (1 << 15) - 1
cdef uint32_t SAMPLE_RATE = 44100


cdef int16_t samp(float v) nogil:
    """Convert a float in [-1, 1] to an int16_t sample."""
    return <int16_t> floor(v * AMPLITUDE)


cdef class Wavetable:
    cdef int16_t[1024] wavetable

    def __init__(self, gen):
        cdef int i
        for i, val in enumerate(gen):
            self.wavetable[i] = samp(val)
            if i == 1023:
                return

        if i != 1023:
            raise ValueError(
                "Wavetable generator generated too few values."
            )

    @staticmethod
    def from_function(f):
        """Generate a wavetable by calling a function f.

        f should take a single float argument between 0 and tau (pi * 2) and
        return values in [-1, 1].
        """
        cdef Wavetable w
        cdef double t

        w = Wavetable.__new__(Wavetable)
        for i in range(1024):
            t = i * pi / 512
            w.wavetable[i] = samp(f(t))
        return w

    @staticmethod
    def sine():
        """Construct a sine waveform."""
        cdef Wavetable w
        cdef size_t i
        w = Wavetable.__new__(Wavetable)
        with nogil:
            for i in range(1024):
                w.wavetable[i] = samp(
                    sin(pi * 2.0 * i / 1023.0)
                )
        return w

    @staticmethod
    def triangle():
        """Construct a triangle waveform."""
        cdef Wavetable w
        cdef size_t i
        cdef float v
        w = Wavetable.__new__(Wavetable)
        with nogil:
            for i in range(1024):
                if i < 256:
                    v = i / 256
                elif i < 768:
                    v = 1.0 - (i - 255) / 256
                else:
                    v = (i - 768) / 256 - 1.0
                w.wavetable[i] = samp(v)
        return w

    @staticmethod
    def saw():
        """Construct a saw waveform."""
        cdef Wavetable w
        cdef size_t i
        cdef float v
        w = Wavetable.__new__(Wavetable)
        with nogil:
            for i in range(1024):
                w.wavetable[(i + 512) % 1024] = samp(i / 512.0 - 1.0)
        return w

    @staticmethod
    def square(float duty_cycle=0.5):
        """Generate a square-wave waveform.

        duty_cycle is the fraction of the period during which the waveform is
        greater than zero.

        """
        cdef Wavetable w
        cdef size_t i, split
        cdef float v
        w = Wavetable.__new__(Wavetable)

        if not (0.0 < duty_cycle < 1.0):
            raise ValueError("duty_cycle must be between 0 and 1")

        split = round(duty_cycle * 1024)
        with nogil:
            for i in range(split):
                w.wavetable[i] = 32767
            for i in range(split, 1024):
                w.wavetable[i] = -32768
        return w

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef Py_ssize_t itemsize = sizeof(int16_t)

        buffer.buf = self.wavetable
        buffer.format = 'h'                     # double
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = sizeof(self.wavetable)
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 0
        buffer.shape = NULL
        buffer.strides = NULL
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass


cdef class SoundBuffer:
    cdef size_t n_samples
    cdef int16_t *samples

    sample_rate: int = SAMPLE_RATE
    channels: int = 1

    def __cinit__(self, size_t n_samples):
        self.samples = <int16_t*> PyMem_Malloc(n_samples * sizeof(int16_t))
        self.n_samples = n_samples
        if not self.samples:
            raise MemoryError()
        for i in range(n_samples):
            self.samples[i] = 0

    def __dealloc__(self):
        PyMem_Free(self.samples)

    def __len__(self):
        return self.n_samples

    def __getitem__(self, ssize_t i):
        if i >= 0:
            if i >= self.n_samples:
                raise IndexError("index out of range")
        else:
            i = self.n_samples + i
            if i < 0:
                raise IndexError("index out of range")
        return self.samples[i]

    @property
    def duration(SoundBuffer self) -> float:
        """Get the duration of this sound in seconds, as a float."""
        return self.n_samples / <float> SAMPLE_RATE

    def save(self, filename: str):
        """Save this sound to a .wav file."""
        import wave

        with wave.open(filename, 'wb') as wav:
            wav.setframerate(SAMPLE_RATE)
            wav.setnchannels(1)
            wav.setnframes(self.n_samples)
            wav.setsampwidth(2)
            wav.writeframesraw(self)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef Py_ssize_t itemsize = sizeof(int16_t)

        buffer.buf = self.samples
        buffer.format = 'h'                     # double
        buffer.internal = NULL                  # see References
        buffer.itemsize = itemsize
        buffer.len = sizeof(int16_t) * self.n_samples
        buffer.ndim = 1
        buffer.obj = self
        buffer.readonly = 0
        buffer.shape = NULL
        buffer.strides = NULL
        buffer.suboffsets = NULL                # for pointer arrays only

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def get_queue_source(self):
        """Duck type as a pyglet.media.Source."""
        return PygletSource(self)


cdef class CachedSound:
    cdef SoundBuffer buf

    def __cinit__(self):
        self.buf = None

    def _clear(self):
        self.buf = None

    def _set(self, SoundBuffer sound):
        self.buf = sound

    def _get(self):
        if self.buf is None:
            self.buf = <SoundBuffer?> self._build()
        return self.buf

    def _build(self) -> SoundBuffer:
        """Override this to build the sound.

        Must return a SoundBuffer.
        """

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        """Delegate to our cached sound.

        If we have no cached sound, attempt to build it by calling build().

        """
        PyObject_GetBuffer(self._get(), buffer, flags)

    def __releasebuffer__(self, Py_buffer *buffer):
        pass


cdef class PygletSource:
    cdef SoundBuffer buf
    cdef size_t pos

    def __init__(self, SoundBuffer buf):
        self.buf = buf
        self.pos = 0

    video_format = None

    @property
    def audio_format(self):
        from pyglet.media.codecs import AudioFormat
        return AudioFormat(
            channels=1,
            sample_size=16,
            sample_rate=self.buf.sample_rate
        )

    def get_audio_data(self, length):
        from pyglet.media.codecs import AudioData
        if self.pos != 0:
            return None

        self.pos = self.buf.n_samples * sizeof(int16_t)
        return AudioData(
            self.buf,
            self.pos,
            self.buf.n_samples / <float> SAMPLE_RATE,
            self.buf.duration,
            ()
        )


def tone(
    Wavetable wavetable,
    double pitch=440.0,  # Hz, default = A
    uint32_t attack=4000,
    uint32_t decay=4000,
    uint32_t sustain=30000,
    uint32_t release=20000
):
    cdef uint64_t time = 0
    cdef size_t n_samples, i
    cdef SoundBuffer t
    cdef int16_t *samples
    cdef uint64_t omega   # angular velocity
    cdef int16_t v
    cdef float amplitude

    # time and omega will be fixed point with a 32-bit fractional part so
    # that we can track time within the sample with simple integer addition.
    #
    # High accuracy is needed because single-bit rounding errors add up
    # over tens of thousands of samples.
    time = 0
    omega = <uint64_t> (pitch * 1024.0 / SAMPLE_RATE * (1 << 32))

    n_samples = attack + decay + sustain + release
    t = SoundBuffer(n_samples)
    samples = t.samples

    with nogil:
        for i in range(n_samples):
            time += omega
            v = wavetable.wavetable[(time >> 32) & 1023]

            if i < attack:
                amplitude = (i / <float> attack)
            elif i < attack + decay:
                amplitude = (1.0 - (i - attack) / <float> decay * 0.3)
            elif i < attack + decay + sustain:
                amplitude = 0.7
            else:
                amplitude = (n_samples - i) / release * 0.7

            samples[i] = <int16_t> (amplitude * v)

    return t


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef void reset_sample(
    float p_base_freq,
    float p_freq_limit,
    float p_freq_ramp,
    float p_freq_dramp,
    float p_duty,
    float p_duty_ramp,
    float p_arp_mod,
    float p_arp_speed,
    double *fperiod,
    int *period,
    double *fmaxperiod,
    double *fslide,
    double *fdslide,
    float *square_duty,
    float *square_slide,
    double *arp_mod,
    int *arp_time,
    int *arp_limit
) nogil:
    fperiod[0] = 100.0 / (p_base_freq * p_base_freq + 0.001)
    period[0] = <int> fperiod[0]
    fmaxperiod[0] = 100.0 / (p_freq_limit * p_freq_limit + 0.001)
    fslide[0] = 1.0 - p_freq_ramp ** 3.0 * 0.01
    fdslide[0] = p_freq_dramp ** 3.0 * -0.000001;
    square_duty[0] = 0.5 - p_duty * 0.5;
    square_slide[0] = p_duty_ramp * -0.00005;
    if p_arp_mod >= 0.0:
        arp_mod[0] = 1.0 - p_arp_mod ** 2.0 * 0.9
    else:
        arp_mod[0] = 1.0 + p_arp_mod ** 2.0 * 10.0
    arp_time[0] = 0
    arp_limit[0] = <int> (((1.0 - p_arp_speed) ** 2.0) * 20000 + 32)
    if p_arp_speed == 1.0:
        arp_limit[0] = 0


from cython cimport floating


cdef void clamp(floating *v, floating min, floating max) nogil:
    """Clamp the given value v to between min and max."""
    if v[0] < min:
        v[0] = min
    elif v[0] > max:
        v[0] = max


@cython.boundscheck(False)
cdef void fill_noise(float *noise_buffer) nogil:
    for i in range(32):
        noise_buffer[i] = frnd(2.0) - 1.0


@cython.cdivision(True)
cdef float frnd(float range_) nogil:
    return <float> (rand() % 10001) / 10000 * range_


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def sfx(
    int wave_type=0,
    float p_base_freq=0.3,
    float p_freq_limit=0.0,
    float p_freq_ramp=0.0,
    float p_freq_dramp=0.0,
    float p_duty=0.0,
    float p_duty_ramp=0.0,
    float p_vib_strength=0.0,
    float p_vib_speed=0.0,
    float p_vib_delay=0.0,
    float p_env_attack=0.0,
    float p_env_sustain=0.3,
    float p_env_decay=0.4,
    float p_env_punch=0.0,
    float p_lpf_resonance=0.0,
    float p_lpf_freq=1.0,
    float p_lpf_ramp=0.0,
    float p_hpf_freq=0.0,
    float p_hpf_ramp=0.0,
    float p_pha_offset=0.0,
    float p_pha_ramp=0.0,
    float p_repeat_speed=0.0,
    float p_arp_speed=0.0,
    float p_arp_mod=0.0,
):
    cdef int phase=0, period, env_stage, env_time, iphase, ipp
    cdef double fperiod, fmaxperiod, fslide, fdslide, arp_mod
    cdef float square_duty, square_slide, env_vol, fphase, fdphase
    cdef int env_length[3]
    cdef float phaser_buffer[1024]
    cdef float noise_buffer[32]
    cdef float fltp = 0.0, fltdp = 0.0, fltphp = 0.0
    cdef float fltw, fltw_d, fltdmp, flthp, flthp_d, vib_phase, vib_speed, vib_amp
    cdef int rep_time, rep_limit, arp_time, arp_limit
    cdef float rfperiod
    cdef float ssample = 0.0, sample = 0.0, fp, pp
    cdef int si, ni

    reset_sample(
        p_base_freq,
        p_freq_limit,
        p_freq_ramp,
        p_freq_dramp,
        p_duty,
        p_duty_ramp,
        p_arp_mod,
        p_arp_speed,

        &fperiod,
        &period,
        &fmaxperiod,
        &fslide,
        &fdslide,
        &square_duty,
        &square_slide,
        &arp_mod,
        &arp_time,
        &arp_limit
    )

    # reset filter
    fltw = 0.1 * p_lpf_freq ** 3.0
    fltw_d = 1.0 + p_lpf_ramp * 0.0001
    fltdmp = 5.0 / (1.0 + p_lpf_resonance ** 2.0 * 20.0) * (0.01 + fltw)
    clamp(&fltdmp, 0.0, 0.8)

    flthp = 0.1 * p_hpf_freq ** 2.0
    flthp_d = 1.0 + p_hpf_ramp * 0.0003;

    # reset vibrato
    vib_phase = 0.0;
    vib_speed = p_vib_speed ** 2.0 * 0.01;
    vib_amp = p_vib_strength * 0.5;

    # reset envelope
    env_vol = 0.0;
    env_stage = 0;
    env_time = 0;
    env_length[0] = <int> (p_env_attack * p_env_attack * 100000.0);
    env_length[1] = <int> (p_env_sustain * p_env_sustain * 100000.0);
    env_length[2] = <int> (p_env_decay * p_env_decay * 100000.0);

    cdef size_t n_samples = env_length[0] + env_length[1] + env_length[2];
    cdef SoundBuffer s = SoundBuffer(n_samples)

    with nogil:
        fphase = p_pha_offset ** 2.0 * 1020.0;
        if p_pha_offset < 0.0:
            fphase = -fphase
        fdphase = p_pha_ramp ** 2.0
        if p_pha_ramp < 0.0:
            fdphase = -fdphase
        iphase = abs(<int> fphase)
        ipp = 0
        for i in range(1024):
            phaser_buffer[i] = 0.0

        # Fill noise buffer
        fill_noise(noise_buffer)

        # reset repeats
        rep_time = 0
        rep_limit = <int> ((1.0 - p_repeat_speed) ** 2.0) * 20000 + 32
        if p_repeat_speed == 0.0:
            rep_limit = 0

        for i in range(n_samples):
            rep_time += 1

            if rep_limit and rep_time >= rep_limit:
                rep_time = 0
                reset_sample(
                    p_base_freq,
                    p_freq_limit,
                    p_freq_ramp,
                    p_freq_dramp,
                    p_duty,
                    p_duty_ramp,
                    p_arp_mod,
                    p_arp_speed,

                    &fperiod,
                    &period,
                    &fmaxperiod,
                    &fslide,
                    &fdslide,
                    &square_duty,
                    &square_slide,
                    &arp_mod,
                    &arp_time,
                    &arp_limit
                )

            # frequency envelopes/arpeggios
            arp_time += 1
            if 0 != arp_limit < arp_time:
                arp_limit = 0
                fperiod *= arp_mod

            fslide += fdslide
            fperiod *= fslide
            if fperiod > fmaxperiod:
                fperiod = fmaxperiod
                if p_freq_limit > 0.0:
                    break

            rfperiod = fperiod
            if vib_amp > 0.0:
                vib_phase += vib_speed
                rfperiod = fperiod * (1.0 + sin(vib_phase) * vib_amp)

            period = <int> rfperiod
            if period < 8:
                period = 8
            square_duty += square_slide
            clamp(&square_duty, 0.0, 0.5)

            # volume envelope
            env_time += 1
            if env_time > env_length[env_stage]:
                env_time = 0
                env_stage += 1
                if env_stage == 3:
                    break

            if env_stage == 0:
                env_vol = <float> env_time / env_length[0]
            elif env_stage == 1:
                # TODO: removed a pow(1.0 - x / y, 1.0) here. Why?
                env_vol = 1.0 + (1.0 - <float> env_time / env_length[1]) * 2.0 * p_env_punch
            elif env_stage == 2:
                env_vol = 1.0 - <float> env_time / env_length[2]

            # phaser step
            fphase += fdphase;
            iphase = abs(<int> fphase)
            if iphase > 1023:
                iphase = 1023

            if flthp_d != 0.0:
                flthp *= flthp_d
                clamp(&flthp, 0.00001, 0.1)

            ssample = 0.0
            for si in range(8):  # 8x supersampling
                sample = 0.0
                phase += 1
                if phase >= period:
                    phase %= period
                    if wave_type == 3:
                        fill_noise(noise_buffer)

                # base waveform
                fp = <float> phase / period;
                if wave_type == 0:  # square
                    sample = 0.5 if fp < square_duty else -0.5
                elif wave_type == 1:  # sawtooth
                    sample = 1.0 - fp * 2
                elif wave_type == 2:  # sine
                    sample = <float> sin(fp * 2 * pi);
                elif wave_type == 3:  # noise
                    sample = noise_buffer[<size_t> (phase * 32 / period)]

                # lp filter
                pp = fltp
                fltw *= fltw_d
                clamp(&fltw, 0.0, 0.1)
                if p_lpf_freq != 1.0:
                    fltdp += (sample - fltp) * fltw;
                    fltdp -= fltdp * fltdmp;
                    fltp += fltdp
                else:
                    fltp = sample
                    fltdp = 0.0

                # hp filter
                fltphp += fltp - pp;
                fltphp -= fltphp * flthp;
                sample = fltphp;

                # phaser
                phaser_buffer[ipp & 1023] = sample
                sample += phaser_buffer[(ipp - iphase + 1024) & 1023]
                ipp = (ipp + 1) & 1023

                # final accumulation and envelope application
                ssample += sample * env_vol

            ssample /= 8
            clamp(&ssample, -1.0, 1.0)
            s.samples[i] = samp(ssample)
    return s


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def pluck(float duration, float pitch, float release=0.1):
    """Generate a pluck sound using the Karplus-Strong algorithm."""
    cdef size_t delay, i, n_samples, release_samples, pos
    cdef int16_t randval, prev
    cdef float fsample
    n_samples = <size_t> (SAMPLE_RATE * duration)
    release_samples = min(<size_t> (SAMPLE_RATE * release), n_samples)

    delay = <size_t> (SAMPLE_RATE / pitch)
    if n_samples < delay:
        raise ValueError(
            f"n_samples must be at least {delay} for pitch {pitch}"
        )

    cdef SoundBuffer s = SoundBuffer(n_samples)

    with nogil:
        prev = 0
        for i in range(delay):
            randval = 1 << 15 if rand() % 2 else -(1 << 15)
            prev = s.samples[i] = (randval >> 1) + (prev >> 1)

        for i in range(delay, n_samples):
            prev = s.samples[i] = (s.samples[i - delay] >> 1) + (prev >> 1)

        # Apply attack and release envelopes
        for pos in range(delay):
            fsample = s.samples[pos] / <float> (1 << 15)
            s.samples[pos] = samp(i * fsample / delay)
        for i in range(release_samples):
            pos = n_samples - i - 1
            fsample = s.samples[pos] / <float> (1 << 15)
            s.samples[pos] = samp(i * fsample / release_samples)

    return s



import re
import math
import random
from functools import lru_cache
from typing import Tuple, Union, Optional, Dict
from enum import Enum

import _pyfxr
from _pyfxr import SoundBuffer, Wavetable, sfx, CachedSound

__all__ = (
    'SAMPLE_RATE',

    'SoundBuffer',
    'Wavetable',

    'SFX',
    'pickup',
    'laser',
    'explosion',
    'powerup',
    'hurt',
    'jump',
    'select',

    'tone',
    'pluck',
    'note_to_hertz',
)


#: The sample rate for all sounds generated by pyfxr
SAMPLE_RATE: int = 44100

NOTE_PATTERN = re.compile(r'^([A-G])([b#]?)([0-8])$')

A4 = 440.0

NOTE_VALUE = dict(C=-9, D=-7, E=-5, F=-4, G=-2, A=0, B=2)

TWELTH_ROOT = math.pow(2, (1 / 12))


@lru_cache()
def note_to_hertz(note: str) -> float:
    """Get a frequency for a given note.

    Here we use an even temper - all semitones in the octave are equally
    spaced twelfth powers of 2.

    """
    note, accidental, octave = validate_note(note)
    value = note_value(note, accidental, octave)
    return A4 * math.pow(TWELTH_ROOT, value)


def note_value(note: str, accidental: str, octave: int) -> int:
    """Given a note name, an accidental and an octave, return a note number.

    Each number corresponds to a specific note and can be converted to a pitch.
    """
    value = NOTE_VALUE[note]
    if accidental:
        value += 1 if accidental == '#' else -1
    return (4 - octave) * -12 + value


def validate_note(note: str) -> Tuple[str, str, int]:
    match = re.match(NOTE_PATTERN, note)
    if match is None:
        raise InvalidNote(
            '%s is not a valid note. '
            'notes are A-F, are either normal, flat (b) or sharp (#) '
            'and of octave 0-8' % note
        )
    note, accidental, octave = match.group(1, 2, 3)
    return note, accidental, int(octave)


def pluck(
    duration: float,
    pitch: Union[float, str],
    release: float = 0.1
) -> SoundBuffer:
    """Generate a pluck sound, like a harp or guitar."""
    # This is a wrapper to handle converting a note string to a pitch
    if isinstance(pitch, str):
        pitch = note_to_hertz(pitch)
    return _pyfxr.pluck(duration, pitch, release)


pluck.__doc__ = _pyfxr.pluck.__doc__


def tone(
    pitch: Union[float, str] = 440.0,  # Hz, default = A
    attack: float = 0.1,
    decay: float = 0.1,
    sustain: float = 0.75,
    release: float = 0.25,
    wavetable: Wavetable = Wavetable.sine(),
) -> SoundBuffer:
    """Generate a tone using a wavetable.

    The tone will be modulated by an ADSR envelope
    (attack-decay-sustain-release) which gives the tone a more natural feel,
    and avoids clicks when played. The total length of the tone is the sum of
    these durations.

    :param wavetable: The wavetable to use (default is a sine wave).
    :param pitch: The pitch of the tone to generate, either float Hz or a note
                  name/number like ``Bb4`` for B-flat in the 4th octave.
    :param attack: Attack time in seconds
    :param decay: Decay time in seconds
    :param sustain: Sustain time in seconds
    :param release: Release time in seconds

    """
    # This is a wrapper to handle converting a note string to a pitch
    if isinstance(pitch, str):
        pitch = note_to_hertz(pitch)
    return _pyfxr.tone(
        wavetable,
        pitch,
        attack * 44100,
        decay * 44100,
        sustain * 44100,
        release * 44100,
    )


class FloatParam:
    """A parameter for a sound effect."""
    name: str
    default: float
    min: Optional[float]
    max: Optional[float]

    def __init__(
        self,
        default: float = 0.0,
        min: Optional[float] = None,
        max: Optional[float] = None
    ):
        self.default = default
        self.min = min
        self.max = max
        if max is not None and min is not None and self.max < self.min:
            raise ValueError("max {self.max} < min {self.min}")
        self.name = None

    def __set_name__(self, cls: type, name: str):
        self.name = name

    def __get__(self, inst: 'SFX', cls: type) -> float:
        return inst._params.get(self.name, self.default)

    def __set__(self, inst: 'SFX', value: float):
        value = float(value)
        if self.min is not None and value < self.min:
            raise ValueError(
                f"{self.name} must be greater than {self.min} (got {value})"
            )
        elif self.max is not None and value > self.max:
            raise ValueError(
                f"{self.name} must be less than {self.max} (got {value})"
            )
        inst._params[self.name] = value
        inst._clear()

    def __delete__(self, inst: 'SFX'):
        inst._params.pop(self.name, None)
        inst._clear()


class WaveType(Enum):
    """The wave types available for the SFX builder.

    Pure tones with tone() use arbitrary wavetables rather than this
    enumeration.

    """
    #: A square-wave waveform
    SQUARE = 0

    #: A saw-wave waveform
    SAW = 1

    #: A sine wave
    SINE = 2

    #: Random noise
    NOISE = 3


class SFX(CachedSound):
    """Build a sound effect using a set of parameters.

    The list of parameters is long and the sensible ranges for the parameters
    aren't that clear. This class acts as a validator and builder for the
    parameters, making it simpler to experiment with sound effects.

    You can also serialise this class in several ways:

    * The ``repr()`` is suitable for pasting into code.
    * You can serialise it as JSON using ``.as_dict()``.
    * You can pickle the class.

    In any of these case the size is much smaller than the generated
    SoundBuffer.

    SFX supports the buffer protocol much like :class:`SoundBuffer`; accessing
    the object as a buffer generates and caches a sound.
    """

    #: The initial frequency of the sound
    base_freq: float = FloatParam(0.3)
    #: The maximum frequency of the sound
    freq_limit: float = FloatParam(0.0)
    #: The rate of change of the frequency of the sound
    freq_ramp: float = FloatParam(0.0)
    #: The acceleration of the change in frequency of the sound
    freq_dramp: float = FloatParam(0.0)

    #: If using square wave, the duty cycle of the waveform
    duty: float = FloatParam(0.0)
    #: The rate of change of the square wave duty cycle
    duty_ramp: float = FloatParam(0.0)

    #: Vibrato strength
    vib_strength: float = FloatParam(0.0)
    #: Vibrato speed
    vib_speed: float = FloatParam(0.0)
    #: Vibrato delay
    vib_delay: float = FloatParam(0.0)

    #: The duration of the attack phase of the ADSR envelope
    env_attack: float = FloatParam(0.0)
    #: The duration of the sustain phase of the ADSR envelope
    env_sustain: float = FloatParam(0.3)
    #: The duration of the decay phase of the ADSR envelope
    env_decay: float = FloatParam(0.4)
    #: Causes the volume to decrease during the sustain phase of the envelope
    env_punch: float = FloatParam(0.0)

    #: Low-pass filter resonance
    lpf_resonance: float = FloatParam(0.0)
    #: Low-pass filter frequency
    lpf_freq: float = FloatParam(1.0)
    #: Low-pass filter ramp
    lpf_ramp: float = FloatParam(0.0)

    #: High-pass filter frequency
    hpf_freq: float = FloatParam(0.0)
    #: High-pass filter ramp
    hpf_ramp: float = FloatParam(0.0)

    #: Phaser offset
    pha_offset: float = FloatParam(0.0)
    #: Phaser ramp
    pha_ramp: float = FloatParam(0.0)

    #: Repeat speed
    repeat_speed: float = FloatParam(0.0)

    #: Arpeggio speed
    arp_speed: float = FloatParam(0.0)
    #: Arpeggio mod
    arp_mod: float = FloatParam(0.0)

    __slots__ = '_params'

    def __init__(self, **kwargs):
        self._params = {}
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def wave_type(self) -> WaveType:
        """Get the wave type."""
        return WaveType(self._params.get('wave_type', 0))

    @wave_type.setter
    def wave_type(self, v: Union[str, int, WaveType]):
        """Set the wave type."""
        if isinstance(v, str):
            v = WaveType[v.upper()]
        self._params['wave_type'] = WaveType(v).value
        self._clear()

    def as_dict(self) -> dict:
        """Get the parameters as a dict.

        The dict is suitable for serialising as JSON; to reconstruct the
        object, pass the parameters as kwargs to the constructor, eg.

        >>> s = SFX(...)
        >>> params = s.as_dict()
        >>> s2 = SFX(**params)

        """
        return self._params.copy()

    def __repr__(self):
        """Generate a repr for this sound effect.

        The repr is guaranteed to be executable code to reconstruct the SFX
        instance.

        """
        params = [f'{type(self).__module__}.{type(self).__qualname__}(']

        for k, desc in vars(type(self)).items():
            if k != 'wave_type' and not isinstance(desc, FloatParam):
                continue

            try:
                value = self._params[k]
            except KeyError:
                continue

            r = repr(value)

            if k == 'wave_type':
                value = WaveType(value)
                r = f'{value.__module__}.{value}'

            params.append(f'    {k}={r},')
        if len(params) == 1:
            return params[0] + ')'
        params.append(')')
        return '\n'.join(params)

    def get_queue_source(self):
        # Duck type as a pyglet.media.Source.
        # Don't put a docstring here as it causes the method to appear in
        # the docs.
        return self._get().get_queue_source()

    def _build(self) -> SoundBuffer:
        """Actually generate the sound using the current parameters."""
        return sfx(**{
            f'p_{k}' if k != 'wave_type' else k: v
            for k, v in self._params.items()
        })

    def build(self) -> SoundBuffer:
        """Get the generated sound (memoised)."""
        return self._get()

    def envelope(
        self,
        attack: float = 0.0,
        sustain: float = 0.3,
        decay: float = 0.4,
        punch: float = 0.0
    ):
        """Set the ADSR envelope for this sound effect."""
        self.env_attack = attack
        self.env_sustain = sustain
        self.env_decay = decay
        self.env_punch = punch
        return self


def one_in(n: int) -> bool:
    """Return True with odds of 1 in n."""
    return not random.randint(0, n)


def _mksfx(params: Dict[str, float]) -> SFX:
    """Round the parameters and construct an SFX.

    Rounding is helpful because it makes the repr shorter and more
    human-friendly.

    """
    return SFX(**{k: round(v, 3) for k, v in params.items()})


def pickup() -> SFX:
    """Generate a random bell sound, like picking up a coin."""
    base_freq = random.uniform(0.4, 0.9)
    env_attack = 0.0
    env_sustain = random.uniform(0.0, 0.1)
    env_decay = random.uniform(0.1, 0.5)
    env_punch = random.uniform(0.3, 0.6)
    if one_in(2):
        arp_mod = random.uniform(0.2, 0.6)
    return _mksfx(locals())


def laser() -> SFX:
    """Generate a random laser sound."""
    wave_type = random.choice((0, 0, 1, 1, 2))
    base_freq = random.uniform(0.5, 1.0)
    freq_limit = max(0.2, base_freq - random.uniform(0.2, 0.8))
    freq_ramp = random.uniform(-0.35, -0.15)
    if one_in(3):
        base_freq = random.uniform(0.3, 0.9)
        freq_limit = random.uniform(0.0, 0.1)
        freq_ramp = random.uniform(-0.65, -0.35)
    if one_in(2):
        duty = random.uniform(0.0, 0.5)
        duty_ramp = random.uniform(0.0, 0.2)
    else:
        duty = random.uniform(0.4, 0.9)
        duty_ramp = random.uniform(-0.7, 0.0)
    env_attack = 0.0
    env_sustain = random.uniform(0.1, 0.3)
    env_decay = random.uniform(0.0, 0.4)
    if one_in(2):
        env_punch = random.uniform(0.0, 0.3)
    if one_in(3):
        pha_offset = random.uniform(0.0, 0.2)
        pha_ramp = random.uniform(-0.2, 0.0)
    if one_in(2):
        hpf_freq = random.uniform(0.0, 0.3)

    return _mksfx(locals())


def explosion() -> SFX:
    """Generate a random explosion sound."""
    wave_type = 3
    if one_in(2):
        base_freq = random.uniform(0.1, 0.5) ** 2
        freq_ramp = random.uniform(-0.1, 0.3)
    else:
        base_freq = random.uniform(0.2, 0.7) ** 2
        freq_ramp = random.uniform(-0.4, -0.2)
    if one_in(5):
        freq_ramp = 0
    if one_in(3):
        repeat_speed = random.uniform(0.3, 0.8)
    env_attack = 0.0
    env_sustain = random.uniform(0.1, 0.4)
    env_decay = random.uniform(0.0, 0.5)
    if one_in(2):
        pha_offset = random.uniform(-0.3, 0.6)
        pha_ramp = random.uniform(-0.3, 0)
    env_punch = random.uniform(0.2, 0.6)
    if one_in(2):
        vib_strength = random.uniform(0.0, 0.7)
        vib_speed = random.uniform(0.0, 0.6)
    if one_in(3):
        arp_speed = random.uniform(0.6, 0.9)
        arp_mod = random.uniform(-0.8, 0.8)

    return _mksfx(locals())


def powerup() -> SFX:
    """Generate a random chime, like receiving a power-up."""
    if one_in(2):
        wave_type = 1
    else:
        duty = random.uniform(0.0, 0.6)

    if one_in(2):
        base_freq = random.uniform(0.2, 0.5)
        freq_ramp = random.uniform(0.1, 0.5)
        repeat_speed = random.uniform(0.4, 0.8)
    else:
        base_freq = random.uniform(0.2, 0.5)
        freq_ramp = random.uniform(0.05, 0.25)
        if one_in(2):
            vib_strength = random.uniform(0.0, 0.7)
            vib_speed = random.uniform(0.0, 0.6)
    env_attack = 0.0
    env_sustain = random.uniform(0.0, 0.4)
    env_decay = random.uniform(0.1, 0.5)

    return _mksfx(locals())


def hurt() -> SFX:
    """Generate a random impact sound, like a character being hurt."""
    wave_type = random.choice([0, 1, 3])
    if wave_type == 0:
        duty = random.uniform(0, 0.6)
    base_freq = random.uniform(0.2, 0.8)
    freq_ramp = random.uniform(-0.7, -0.3)
    env_attack = 0.0
    env_sustain = random.uniform(0.0, 0.1)
    env_decay = random.uniform(0.1, 0.3)
    if one_in(2):
        hpf_freq = random.uniform(0.0, 0.3)

    return _mksfx(locals())


def jump() -> SFX:
    """Generate a random jump sound."""
    wave_type = 0
    duty = random.uniform(0.0, 0.6)
    base_freq = random.uniform(0.3, 0.6)
    freq_ramp = random.uniform(0.1, 0.3)
    env_attack = 0.0
    env_sustain = random.uniform(0.1, 0.4)
    env_decay = random.uniform(0.1, 0.3)
    if one_in(2):
        hpf_freq = random.uniform(0.0, 0.3)
    if one_in(2):
        lpf_freq = random.uniform(0.4, 1.0)

    return _mksfx(locals())


def select() -> SFX:
    """Generate a random 'blip' noise, like selecting an option in a menu."""
    wave_type = random.choice([0, 1])
    if wave_type == 0:
        duty = random.uniform(0.0, 0.6)
    base_freq = random.uniform(0.2, 0.6)
    env_attack = 0.0
    env_sustain = random.uniform(0.1, 0.2)
    env_decay = random.uniform(0.0, 0.2)
    hpf_freq = 0.1
    return _mksfx(locals())
